@@ -3,14 +3,14 @@
  *
  * Copyright 1999-2004 Gentoo Foundation
  * Distributed under the terms of the GNU General Public License v2
- * $Header: /var/cvsroot/gentoo-src/sandbox/localdecls.h,v 1.2 2005/04/23 20:44:15 azarah Exp $
+ * $Header: /var/cvsroot/gentoo-src/sandbox/localdecls.h,v 1.3 2005/05/02 23:53:52 vapier Exp $
  */
 
 #ifndef __LOCALDECLS_H__
 #define __LOCALDECLS_H__
 
 /* take care of broken ld loading */
-#if defined(__GLIBC__)
+#if defined(__GLIBC__) && !defined(__UCLIBC__)
 
 # if __GLIBC__ <= 2 && __GLIBC_MINOR__ <= 2
 #  define BROKEN_RTLD_NEXT
