@@ -11,7 +11,7 @@
 **    Copyright (C) 2001 Geert Bevin, Uwyn, http://www.uwyn.com
 **    Distributed under the terms of the GNU General Public License, v2 or later 
 **    Author : Geert Bevin <gbevin@uwyn.com>
-**  $Header: /var/cvsroot/gentoo-src/sandbox/sandbox.c,v 1.4 2005/03/13 23:23:00 azarah Exp $
+**  $Header: /var/cvsroot/gentoo-src/sandbox/sandbox.c,v 1.7 2005/05/04 14:46:50 azarah Exp $
 */
 
 /* #define _GNU_SOURCE */
@@ -301,7 +301,7 @@ void setenv_sandbox_write(char *home_dir, char *portage_tmp_dir, char *var_tmp_d
 		snprintf(buf, sizeof(buf),
 			 "%s:%s/.gconfd/lock:%s/.bash_history:",
 			 "/dev/zero:/dev/fd/:/dev/null:/dev/pts/:"
-			 "/dev/vc/:/dev/tty:/tmp/:"
+			 "/dev/vc/:/dev/pty:/dev/tty:/tmp/:"
 			 "/dev/shm/ngpt:/var/log/scrollkeeper.log:"
 			 "/usr/tmp/conftest:/usr/lib/conftest:"
 			 "/usr/lib32/conftest:/usr/lib64/conftest:"
@@ -668,7 +668,22 @@ int main(int argc, char **argv)
 
 		setenv(ENV_SANDBOX_DIR, sandbox_dir, 1);
 		setenv(ENV_SANDBOX_LIB, sandbox_lib, 1);
-		setenv("LD_PRELOAD", sandbox_lib, 1);
+		setenv(ENV_SANDBOX_BASHRC, sandbox_rc, 1);
+		if (NULL != getenv("LD_PRELOAD")) {
+			tmp_string = malloc(strlen(getenv("LD_PRELOAD")) +
+					strlen(sandbox_lib) + 2);
+			if (NULL == tmp_string) {
+				perror(">>> Out of memory (LD_PRELOAD)");
+				exit(1);
+			}
+			strncpy(tmp_string, sandbox_lib, sizeof(sandbox_lib));
+			strncat(tmp_string, " ", 1);
+			strncat(tmp_string, getenv("LD_PRELOAD"), strlen(getenv("LD_PRELOAD")));
+			setenv("LD_PRELOAD", tmp_string, 1);
+			free(tmp_string);
+		} else {
+			setenv("LD_PRELOAD", sandbox_lib, 1);
+		}
 
 		if (!getenv(ENV_SANDBOX_DENY))
 			setenv(ENV_SANDBOX_DENY, LD_PRELOAD_FILE, 1);
